#ifndef BTIMER_HPP
#define BTIMER_HPP

#include <OS.h>
#include <Handler.h>
#include <Message.h>
#include <Messenger.h>

int32 timer_thread_handler(void *data);

class BTimerList;

class BTimer {
private:
	thread_id timer_thread;
	bigtime_t interval;
	bool running;
	
	sem_id pause_sem;
	sem_id lock_sem;
	
	friend int32 timer_thread_handler(void *data);
	
	void ThreadHandler(void);
	
	//Used for BTimerList's automatic timer deletion
	friend class BTimerList;
	BTimer **head;
	BTimer *prev, *next;
	
public:
	BTimer(bigtime_t usec);
	virtual ~BTimer();
	
	virtual void OnInterval(void);
	
	void Resume(void);
	void Pause(void);
	bool Running(void);
	
	void SetInterval(bigtime_t usec);
	bigtime_t Interval(void);
};

class BTimerList {
private:
	BTimer *head;
	
public:
	BTimerList() {
		head = NULL;
	}
	void DeleteAllTimers(void) {
		while (head) {
			delete(head);
		}
	}
	void PushTimer(BTimer *t) {
		if (head) {
			t->prev = head->prev;
			t->next = head;
			t->prev->next = t;
			t->next->prev = t;
			head = t;
		} else {
			head = t;
		}
		t->head = &head;
	}
	~BTimerList() {
		DeleteAllTimers();
	}
};

template<class T>
class BTimerCB : public BTimer {
private:
	T *parent;
	void (T::*callback)(void *data);
	void *data;
	
public:
	BTimerCB(T *parent, void (T::*callback)(void *data), bigtime_t usec,
		void *data = NULL, bool start = true)
		: BTimer(usec)
	{
		this->parent = parent;
		this->callback = callback;
		this->data = data;
		
		if (start)
			Resume();
	}
	
	virtual void OnInterval(void) {
		(parent->*callback)(data);
	}
};

class BTimerMsg : public BTimer {
private:
	BHandler *parent;
	BMessage *message;

public:
	BTimerMsg(BHandler *parent, BMessage *message, bigtime_t usec,
		bool start = true)
		: BTimer(usec)
	{
		this->parent = parent;
		this->message = message;
		
		if (start)
			Resume();
	}
	
	~BTimerMsg() {
		delete message;
	}
	
	virtual void OnInterval(void) {
		BMessenger messenger(parent);
		messenger.SendMessage(message);
	}
};

template<class T>
class BTimerMixin : public BTimerList {
public:
	
	BTimer *CreateTimer(BMessage *message, bigtime_t usec,
		bool start = true)
	{
		BTimer *ret = new BTimerMsg((T*)this, message, usec, start);
		
		PushTimer(ret);
		return ret;
	}
	
	BTimer *CreateTimer(uint32 command, bigtime_t usec,
		bool start = true)
	{
		BMessage *message = new BMessage(command);
		return CreateTimer(message, usec, start);
	}
	
	BTimer *CreateCallbackTimer(void (T::*callback)(void *data), bigtime_t usec,
		void *data = NULL, bool start = true)
	{
		BTimer *ret = new BTimerCB<T> ((T*)this, callback, usec, data, start);
			//TODO:  Use a new-style cast here (I don't know which one's appropriate)
		
		PushTimer(ret);
		return ret;
	}
	
};

#endif
