
#include <Application.h>
#include <Window.h>
#include <Locker.h>

#include <stdio.h>

#include "BTimer.hpp"

#define POLL_WINDOW_SIZE 'pwsz'

struct Ball {
	float x, y;
	float radius;
	
	float dx, dy;
	
	uchar r, g, b, a;
	
	Ball() {
		x = y = 0.0;
		radius = 24.0;
		
		dx = dy = 0;
	}
	
	Ball(float x, float y, float radius, float dx, float dy, uchar r, uchar g, uchar b, uchar a) {
		this->x = x;
		this->y = y;
		this->radius = radius;
		
		this->dx = dx;
		this->dy = dy;
		
		this->r = r;
		this->g = g;
		this->b = b;
		this->a = a;
	}
	
	void Update(float w, float h) {
		x += dx;
		y += dy;
		
		if (x-radius < 0 && dx < 0)
			dx = -dx;
		else if (x+radius >= w && dx > 0)
			dx = -dx;
		
		if (x-radius < 0)
			x = radius;
		else if (x+radius >= w)
			x = w-radius;
		
		if (y-radius < 0 && dy < 0)
			dy = -dy;
		else if (y+radius >= h && dy > 0)
			dy = -dy;
		
		if (y-radius < 0)
			y = radius;
		else if (y+radius >= h)
			y = h-radius;
	}
};

class View : public BView, public BTimerMixin<View>
{
private:
	
	BLocker game_lock;
	Ball balls[3];
	float width, height;
	
	void DrawBall(Ball &ball) {
		SetDrawingMode(B_OP_ALPHA);
		
		SetLowColor(ball.r, ball.g, ball.b, ball.a);
		
		FillEllipse(BRect(
			ball.x-ball.radius, ball.y-ball.radius,
			ball.x+ball.radius, ball.y+ball.radius
		), B_SOLID_LOW);
		StrokeEllipse(BRect(
			ball.x-ball.radius, ball.y-ball.radius,
			ball.x+ball.radius, ball.y+ball.radius
		), B_SOLID_HIGH);
	}
	
public:
	
	View(BRect frame, const char *name, uint32 resizingMode)
		: BView(frame, name, resizingMode, B_WILL_DRAW)
	{
		balls[0] = Ball(10, 20, 30, 1.0, 2.0, 255,0,0,255);
		balls[1] = Ball(30, 70, 20, 5.0, 5.0, 0,255,0,128);
		balls[2] = Ball(200, 80, 40, -10.0, 5.0, 0,0,255,64);
		
		width = Bounds().Width()+1;
		height = Bounds().Height()+1;
	}
	
protected:
	
	void UpdateBall(void *data) {
		Ball *ball = (Ball*)data;
		game_lock.Lock();
		ball->Update(width, height);
		game_lock.Unlock();
	}
	
	void RedrawCallback(void *) {
		/* Careful: Callback timers call from a different thread, so we have to
		   lock the looper before messing with the BView.  Moreover, the BView
		   might get destroyed during the LockLooper, so we check the status to
		   make sure that didn't happen. */
		
		if (!LockLooper())
			return;
		
		Invalidate();
		
		UnlockLooper();
	}
	
	void AttachedToWindow() {
		CreateCallbackTimer(&View::UpdateBall, 15000, &balls[0]);
		CreateCallbackTimer(&View::UpdateBall, 60000, &balls[1]);
		CreateCallbackTimer(&View::UpdateBall, 500000, &balls[2]);
		
		CreateCallbackTimer(&View::RedrawCallback, 15000);
		
		CreateTimer(POLL_WINDOW_SIZE, 200000);
	}
	
	void Draw(BRect updateRect) {
		(void) updateRect;
		
		game_lock.Lock();
		DrawBall(balls[0]);
		DrawBall(balls[1]);
		DrawBall(balls[2]);
		game_lock.Unlock();
	}
	
	void MessageReceived(BMessage *msg) {
		switch (msg->what) {
			case POLL_WINDOW_SIZE:
				game_lock.Lock();
				width = Bounds().Width()+1;
				height = Bounds().Height()+1;
				game_lock.Unlock();
				break;
			default:
				BView::MessageReceived(msg);
		}
	}
};

class Window : public BWindow
{
private:
	
	View *view;

public:
	
	Window(BRect frame) :
		BWindow(frame, "BTimer Sample", B_TITLED_WINDOW, 0)
	{
		view = new View(Bounds().InsetByCopy(0, 0).OffsetToSelf(0, 0),
			"View", B_FOLLOW_ALL_SIDES);
		view->SetViewColor(0x99, 0xCC, 0xFF);
		AddChild(view);
	}
	
	bool QuitRequested() {
		be_app_messenger.SendMessage(B_QUIT_REQUESTED);
		return BWindow::QuitRequested();
	}
};

class Application : public BApplication
{
	public:
	
	Application() : BApplication("application/x-vnd.BTimerSample")
	{
		
	}
	
	void ReadyToRun() {
		BWindow *win = new Window(BRect(100,100,500,400));
		win->Show();
	}
};

//trivial example
class Counter : public BTimerMixin<Counter>
{
private:
	int count;
	
public:
	void Callback(void*) {
		printf("%d\n", count++);
	}
	
	Counter() {
		count = 0;
		CreateCallbackTimer(&Counter::Callback, 1000000);
	}
};

int main(void) {
	Application app;
	Counter counter;
	
	app.Run();
	
	return 0;
}
